from netgen.geom2d import SplineGeometry
from netgen.meshing import MeshingParameters
from ngsolve import *
from xfem import *
# For LevelSetAdaptationMachinery
from xfem.lsetcurv import *
  
def power(u,p):
    if p == 0: 
        return 1
    else:
        return u * power(u,p-1)

# normal derivative jumps:
def dnjump(u,order,comp = -1):
    if order%2==0:
        return dn(u,order,comp) - dn(u.Other(),order,comp)
    else:
        return dn(u,order,comp) + dn(u.Other(),order,comp)

square = SplineGeometry()
square.AddRectangle([0,0], [1,1], bc=1) # Background mesh
maxh= 0.8
ngmesh = square.GenerateMesh(maxh=maxh, quad_dominated=False)
mesh = Mesh(ngmesh) 
order=1
# stabilization parameter for ghost-penalty
gamma_stab = [0.1,0.01,0.001,0.0001,0.00001,0.00001]

levelset = (sqrt((x-0.5)*(x-0.5)+(y-0.5)*(y-0.5)) - 0.4) 
lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=0.1, discontinuous_qn=True)
deformation = lsetmeshadap.CalcDeformation(levelset)
lsetp1 = lsetmeshadap.lset_p1

lset_neg = { "levelset" : lsetp1, "domain_type" : NEG, "subdivlvl" : 0}
lset_if  = { "levelset" : lsetp1, "domain_type" : IF , "subdivlvl" : 0}

ci = CutInfo(mesh,lsetp1)
hasneg = BitArray(ci.GetElementsOfType(NEG))
hasneg |= ci.GetElementsOfType(IF)
hasif = BitArray(ci.GetElementsOfType(IF))
hasif |= ci.GetElementsOfType(IF)
ba_facets = GetFacetsWithNeighborTypes(mesh,a=hasneg,b=hasif)
# facets on which ghost penalty stabilization should be applied
cf_ghost = IndicatorCF(mesh,ba_facets,facets=True)
h = specialcf.mesh_size  

V =  H1(mesh, order = order, dirichlet=[],flags = {"dgjumps":True})
u,v = V.TnT()

u,v = V.TrialFunction(), V.TestFunction()
 
# ghost penalty terms:
gp_term = CoefficientFunction(0.0)
for i in range(order):
    gp_term += gamma_stab[i] * power(h,2*i+1) * dnjump(u,i+1)*dnjump(v,i+1)

#a = BilinearForm(V,symmetric=False)
#a += SymbolicBFI(form = cf_ghost*gp_term,VOL_or_BND = VOL, skeleton=True)
#a.Assemble()

import scipy.sparse as sp
dS = dx(element_boundary=True)
bfi = 0.5*cf_ghost* gp_term *dS
b = BilinearForm(V,symmetric=False) 
b +=bfi
b.Assemble()
rows, cols, vals = b.mat.COO()
B=sp.csr_matrix((vals,(rows, cols)))
#print("B", B.toarray())
#input("B")
########################## 
list_1=[]
for el in V.Elements(VOL):
    list_1.append(el.dofs)

for i in range(V.ndof):
    for j in range(V.ndof):
        print("Entry (i,j) i:", i, "j:", j)
        summands=list()
        for k in range(len(list_1)):
            if (i in list_1[k] and j in list_1[k]):
                t1=list_1[k].index(i) 
                t2=list_1[k].index(j) 
                elid = ElementId(k)
                # the element provides the shape functions
                el = V.GetFE(elid)
  
                # the mapping from reference element to physical element
                trafo = mesh.GetTrafo(elid)

                dofs = V.GetDofNrs(elid)
                elmat = bfi.CalcElementMatrix(el, trafo) 
                print("elmat[t1,t2]:", elmat[t1,t2])
                print ("element matrix:", elmat) 
                summands.append(elmat[t1,t2])
        entry=sum(summands)
        print("Entry (i,j) i:", i, "j:",  j, "Β(i,j):", Β[i,j], "entry:", entry)
        input("NEXT ENTRY") 

